use crate::smallint::{SmallIntType, SmallUintType};
use crate::{SmallInt, SmallUint};
use core::cmp::Ordering;
use core::mem::ManuallyDrop;
use core::ops::{BitAnd, BitOr, BitXor};
use core::ops::{BitAndAssign, BitOrAssign, BitXorAssign};

macro_rules! logic_op {
    (
        $imp:ident, $imp_assign:ident, $typ:ident, $typ_inner:ident, $fun:ident, $fun_assign:ident;
        $i:ident, $j:ident, $p:ident, $q:ident, $s:ident, $t:ident;
        $inline_heap:tt, $heap_heap:tt
    ) => {
        impl<'a, 'b> $imp<&'a $typ> for &'b $typ {
            type Output = $typ;

            fn $fun(self, rhs: &$typ) -> Self::Output {
                match (&self.0, &rhs.0) {
                    (&$typ_inner::Inline($i), &$typ_inner::Inline($j)) => {
                        $typ($typ_inner::Inline($i.$fun($j)))
                    }

                    (&$typ_inner::Inline($i), &$typ_inner::Heap(($p, $s)))
                    | (&$typ_inner::Heap(($p, $s)), &$typ_inner::Inline($i)) => {
                        $inline_heap
                    }

                    (&$typ_inner::Heap(($p, $s)), &$typ_inner::Heap(($q, $t))) => {
                        $heap_heap
                    }
                }
            }
        }

        impl<'a> $imp<$typ> for &'a $typ {
            type Output = $typ;

            fn $fun(self, rhs: $typ) -> Self::Output {
                self.$fun(&rhs)
            }
        }

        impl<'a> $imp<&'a $typ> for $typ {
            type Output = $typ;

            fn $fun(self, rhs: &$typ) -> Self::Output {
                (&self).$fun(rhs)
            }
        }

        impl $imp<$typ> for $typ {
            type Output = $typ;

            fn $fun(self, rhs: $typ) -> Self::Output {
                (&self).$fun(&rhs)
            }
        }

        impl<'a> $imp_assign<&'a $typ> for $typ {
            fn $fun_assign(&mut self, rhs: &'a $typ) {
                *self = (&*self).$fun(rhs);
            }
        }

        impl $imp_assign<$typ> for $typ {
            fn $fun_assign(&mut self, rhs: $typ) {
                *self = (&*self).$fun(rhs);
            }
        }
    };
    (
        $imp:ident, $imp_assign:ident, $typ:ident, $typ_inner:ident, $fun:ident, $fun_assign:ident;
        $i:ident, $j:ident, $p:ident, $q:ident, $s:ident, $t:ident, $slice:ident, $slice1:ident, $slice2:ident, $min:ident, $res:ident;
        $inline_heap_inner:tt, $heap_heap_create_res:tt, $heap_heap_return:tt
    ) => {
        logic_op! {
            $imp, $imp_assign, $typ, $typ_inner, $fun, $fun_assign;
            $i, $j, $p, $q, $s, $t;
            {
                let $slice = unsafe { core::slice::from_raw_parts($p, $s) };

                $inline_heap_inner
            },
            {
                let $slice1 = unsafe { core::slice::from_raw_parts($p, $s) };
                let $slice2 = unsafe { core::slice::from_raw_parts($q, $t) };

                let $min = std::cmp::min($slice1.len(), $slice2.len());

                #[allow(unused_mut)]
                let mut $res = $heap_heap_create_res;

                $heap_heap_return
            }
        }
    };
}


macro_rules! inline_heap_to_inline {
    ($fun:ident, $typ:ident, $typ_inner:ident, $typ_inline:ident; $i:ident, $slice:ident) => {
        let mut j = 0;
        for i in 0..4 {
            j <<= 32;
            j |= $slice[3 - i] as $typ_inline;
        }
        $typ($typ_inner::Inline($i.$fun(j)))
    }
}

macro_rules! inline_heap_to_heap {
    ($fun_assign:ident, $typ:ident, $typ_inner:ident; $i:ident, $slice:ident) => {
        let mut retvec = $slice.to_vec();

        let mut v = $i;
        #[allow(clippy::needless_range_loop)]
        for r in 0..4 {
            retvec[r].$fun_assign(v as u32);

            v >>= 32;
        }

        let mut retslice = ManuallyDrop::new(retvec.into_boxed_slice());

        $typ($typ_inner::Heap((retslice.as_mut_ptr(), retslice.len())))
    }
}


macro_rules! heap_heap_create_res_shortest {
    ($fun:ident; $min:ident, $slice1:ident, $slice2:ident) => {
        let mut res = Vec::with_capacity($min);
        for l in 0..$min {
            res.push($slice1[l].$fun($slice2[l]));
        }
        res
    }
}

macro_rules! heap_heap_create_res_longest {
    ($fun:ident; $slice1:ident, $slice2:ident, $min:ident) => {
        let mut res = if $slice1.len() > $slice2.len() {
            $slice1.to_vec()
        } else {
            $slice2.to_vec()
        };
        for l in 0..$min {
            res[l] = $slice1[l].$fun($slice2[l]);
        }
        res
    }
}


macro_rules! heap_heap_return_heap {
    ($typ:ident, $typ_inner:ident; $res:ident) => {
        let mut slice = ManuallyDrop::new($res.into_boxed_slice());
        $typ($typ_inner::Heap((slice.as_mut_ptr(), slice.len().try_into().unwrap())))
    }
}

macro_rules! heap_heap_return_any {
    ($typ:ident, $typ_inner:ident, $typ_inline:ident; $res:ident) => {
        while $res.len() != 1 && $res[$res.len() - 1] == 0 {
            $res.pop();
        }

        if $res.len() <= 4 {
            let mut r = 0;
            for t in 0..$res.len() {
                r <<= 32;
                r |= $res[$res.len() - 1 - t] as $typ_inline;
            }
            $typ($typ_inner::Inline(r))
        } else {
            let mut slice = ManuallyDrop::new($res.into_boxed_slice());
            $typ($typ_inner::Heap((slice.as_mut_ptr(), slice.len().try_into().unwrap())))
        }
    }
}

logic_op! {
    BitAnd, BitAndAssign, SmallUint, SmallUintType, bitand, bitand_assign;
    i, j, p, q, s, t, slice, slice1, slice2, min, res;
    { inline_heap_to_inline! { bitand, SmallUint, SmallUintType, u128; i, slice } },
    { heap_heap_create_res_shortest! { bitand; min, slice1, slice2 } },
    { heap_heap_return_any! { SmallUint, SmallUintType, u128; res } }
}

logic_op! {
    BitOr, BitOrAssign, SmallUint, SmallUintType, bitor, bitor_assign;
    i, j, p, q, s, t, slice, slice1, slice2, min, res;
    { inline_heap_to_heap! { bitor_assign, SmallUint, SmallUintType; i, slice } },
    { heap_heap_create_res_longest! { bitor; slice1, slice2, min }},
    { heap_heap_return_heap! { SmallUint, SmallUintType; res } }
}

logic_op! {
    BitXor, BitXorAssign, SmallUint, SmallUintType, bitxor, bitxor_assign;
    i, j, p, q, s, t, slice, slice1, slice2, min, res;
    { inline_heap_to_heap! { bitxor_assign, SmallUint, SmallUintType; i, slice } },
    { heap_heap_create_res_longest! { bitxor; slice1, slice2, min }},
    { heap_heap_return_any! { SmallUint, SmallUintType, u128; res } }
}




logic_op! {
    BitXor, BitXorAssign, SmallInt, SmallIntType, bitxor, bitxor_assign;
    i, j, p, q, s, t;
    {
        todo!()
    },
    {
        let slice1 = unsafe { core::slice::from_raw_parts(p, s.unsigned_abs()) };
        let slice2 = unsafe { core::slice::from_raw_parts(q, t as usize) };
        match (s.cmp(&0), t.cmp(&0)) {
            (Ordering::Greater, Ordering::Greater) => {
                let min = std::cmp::min(slice1.len(), slice2.len());

                #[allow(unused_mut)]
                let mut res = { heap_heap_create_res_longest! { bitxor; slice1, slice2, min } };

                heap_heap_return_any! { SmallInt, SmallIntType, i128; res }
            },
            (Ordering::Greater, Ordering::Less) => todo!(),
            (Ordering::Less, Ordering::Greater) => todo!(),
            (Ordering::Less, Ordering::Less) => todo!(),
            (Ordering::Equal, _) | (_, Ordering::Equal) => unreachable!("0 must be inline"),
        }
    }
}
